import re
from typing import Union, TYPE_CHECKING, Any

from Crypto.Hash import SHA1
from crytic_compile.utils.naming import Filename
from slither.core.context.context import Context
from slither.exceptions import SlitherException

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit


# We split the source mapping into two objects
# The reasoning is to allow any object to just inherit from SourceMapping
# To have then everything accessible through obj.source_mapping._
# All an object needs to do is to inherits from SourceMapping
# And call set_offset at some point


class Source:
    def __init__(self, compilation_unit: "SlitherCompilationUnit") -> None:
        self.start: int = 0
        self.length: int = 0
        self.filename: Filename = Filename("", "", "", "")
        self.is_dependency: bool = False
        self.lines: list[int] = []
        self.starting_column: int = 0
        self.ending_column: int = 0
        self.end: int = 0
        self.compilation_unit = compilation_unit

    def to_json(self) -> dict:
        return {
            "start": self.start,
            "length": self.length,
            # TODO investigate filename_used usecase
            # It creates non-deterministic result
            # As it sometimes refer to the relative and sometimes to the absolute
            # "filename_used": self.filename.used,
            "filename_relative": self.filename.relative,
            "filename_absolute": self.filename.absolute,
            "filename_short": self.filename.short,
            "is_dependency": self.is_dependency,
            "lines": self.lines,
            "starting_column": self.starting_column,
            "ending_column": self.ending_column,
        }

    def to_markdown(self, markdown_root: str) -> str:
        lines = self._get_lines_str(line_descr="L")
        filename_relative: str = self.filename.relative if self.filename.relative else ""
        return f"{markdown_root}{filename_relative}{lines}"

    def to_detailed_str(self) -> str:
        lines = self._get_lines_str()
        filename_short: str = self.filename.short if self.filename.short else ""
        return f"{filename_short}{lines} ({self.starting_column} - {self.ending_column})"

    def _get_lines_str(self, line_descr: str = "") -> str:
        line_prefix = self.compilation_unit.core.line_prefix

        lines = self.lines
        if not lines:
            return ""
        if len(lines) == 1:
            return f"{line_prefix}{line_descr}{lines[0]}"

        return f"{line_prefix}{line_descr}{lines[0]}-{line_descr}{lines[-1]}"

    @property
    def content(self) -> str:
        """
        Return the txt content of the Source

        Use this property instead of eg source_code[start:end]
        Above will return incorrect content if source_code contains any unicode
        because self.start and self.end are byte offsets, not char offsets

        Returns: str
        """
        # If the compilation unit was not initialized, it means that the set_offset was never called
        # on the corresponding object, which should not happen
        assert self.compilation_unit
        return (
            self.compilation_unit.core.source_code[self.filename.absolute]
            .encode("utf8")[self.start : self.end]
            .decode("utf8")
        )

    @property
    def content_hash(self) -> str:
        """
        Return sha1(self.content)

        Returns:

        """
        h = SHA1.new()
        h.update(self.content.encode("utf8"))
        return h.hexdigest()

    def __str__(self) -> str:
        lines = self._get_lines_str()
        filename_short: str = self.filename.short if self.filename.short else ""
        return f"{filename_short}{lines}"

    def __hash__(self) -> int:
        return hash(
            (
                self.start,
                self.length,
                self.filename.relative,
                self.end,
            )
        )

    def __eq__(self, other: Any) -> bool:
        try:
            return (
                self.start == other.start
                and self.filename.relative == other.filename.relative
                and self.is_dependency == other.is_dependency
                and self.end == other.end
            )
        except AttributeError:
            return NotImplemented


def _compute_line(
    compilation_unit: "SlitherCompilationUnit", filename: Filename, start: int, length: int
) -> tuple[list[int], int, int]:
    """
    Compute line(s) numbers and starting/ending columns
    from a start/end offset. All numbers start from 1.

    Not done in an efficient way
    """

    start_line, starting_column = compilation_unit.core.crytic_compile.get_line_from_offset(
        filename, start
    )
    try:
        end_line, ending_column = compilation_unit.core.crytic_compile.get_line_from_offset(
            filename, start + length
        )
    except KeyError:
        # This error may occur when the build is not synchronised with the source code on disk.
        # See the GitHub issue https://github.com/crytic/slither/issues/2296
        msg = f"""The source code appears to be out of sync with the build artifacts on disk.
        This discrepancy can occur after recent modifications to {filename.short}. To resolve this
        issue, consider executing the clean command of the build system (e.g. forge clean).
        """
        # We still re-raise the exception as a SlitherException here
        raise SlitherException(msg) from None

    return list(range(start_line, end_line + 1)), starting_column, ending_column


def _convert_source_mapping(offset: str, compilation_unit: "SlitherCompilationUnit") -> Source:
    """
    Convert a text offset to a real offset
    see https://solidity.readthedocs.io/en/develop/miscellaneous.html#source-mappings
    Returns:
        (dict): {'start':0, 'length':0, 'filename': 'file.sol'}
    """
    sourceUnits = compilation_unit.source_units

    position = re.findall("([0-9]*):([0-9]*):([-]?[0-9]*)", offset)
    if len(position) != 1:
        return Source(compilation_unit)

    s, l, f = position[0]
    s = int(s)
    l = int(l)
    f = int(f)

    if f not in sourceUnits:
        new_source = Source(compilation_unit)
        new_source.start = s
        new_source.length = l
        return new_source
    filename_used = sourceUnits[f]

    # If possible, convert the filename to its absolute/relative version
    assert compilation_unit.core.crytic_compile

    filename: Filename = compilation_unit.core.crytic_compile.filename_lookup(filename_used)
    is_dependency = compilation_unit.core.crytic_compile.is_dependency(filename.absolute)

    (lines, starting_column, ending_column) = _compute_line(compilation_unit, filename, s, l)

    new_source = Source(compilation_unit)
    new_source.start = s
    new_source.length = l
    new_source.filename = filename
    new_source.is_dependency = is_dependency
    new_source.lines = lines
    new_source.starting_column = starting_column
    new_source.ending_column = ending_column
    new_source.end = new_source.start + l

    return new_source


class SourceMapping(Context):
    def __init__(self) -> None:
        super().__init__()
        self.source_mapping: Source | None = None
        self.references: list[Source] = []

        self._pattern: str | None = None

    def set_offset(
        self, offset: Union["Source", str], compilation_unit: "SlitherCompilationUnit"
    ) -> None:
        assert compilation_unit
        if isinstance(offset, Source):
            self.source_mapping = offset
        else:
            self.source_mapping = _convert_source_mapping(offset, compilation_unit)
        self.source_mapping.compilation_unit = compilation_unit

    def add_reference_from_raw_source(
        self, offset: str, compilation_unit: "SlitherCompilationUnit"
    ) -> None:
        s = _convert_source_mapping(offset, compilation_unit)
        self.references.append(s)

    @property
    def pattern(self) -> str:
        if self._pattern is None:
            # Add " " to look after the first solidity keyword
            return f" {self.name}"

        return self._pattern



"""
Module detecting the incorrect use of unary expressions
"""

from slither.core.expressions.assignment_operation import AssignmentOperation
from slither.core.expressions.expression import Expression
from slither.core.expressions.unary_operation import UnaryOperationType, UnaryOperation
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output
from slither.visitors.expression.expression import ExpressionVisitor


class InvalidUnaryExpressionDetector(ExpressionVisitor):
    def __init__(self, expression: Expression) -> None:
        self.result: bool = False
        super().__init__(expression)

    def _post_assignement_operation(self, expression: AssignmentOperation) -> None:
        if isinstance(expression.expression_right, UnaryOperation):
            if expression.expression_right.type == UnaryOperationType.PLUS_PRE:
                self.result = True


class InvalidUnaryStateVariableDetector(ExpressionVisitor):
    def __init__(self, expression: Expression) -> None:
        self.result: bool = False
        super().__init__(expression)

    def _post_unary_operation(self, expression: UnaryOperation) -> None:
        if expression.type == UnaryOperationType.PLUS_PRE:
            self.result = True


class IncorrectUnaryExpressionDetection(AbstractDetector):
    """
    Incorrect Unary Expression detector
    """

    ARGUMENT = "incorrect-unary"
    HELP = "Dangerous unary expressions"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#dangerous-unary-expressions"
    )

    WIKI_TITLE = "Dangerous unary expressions"
    WIKI_DESCRIPTION = "Unary expressions such as `x=+1` probably typos."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```Solidity
contract Bug{
    uint public counter;

    function increase() public returns(uint){
        counter=+1;
        return counter;
    }
}
```
`increase()` uses `=+` instead of `+=`, so `counter` will never exceed 1."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove the unary expression."

    def _detect(self) -> list[Output]:
        """
        Detect the incorrect use of unary expressions
        """
        results = []
        for c in self.contracts:
            for variable in c.state_variables:
                if (
                    variable.expression
                    and InvalidUnaryStateVariableDetector(variable.expression).result
                ):
                    info: DETECTOR_INFO = [
                        variable,
                        f" uses an dangerous unary operator: {variable.expression}\n",
                    ]
                    json = self.generate_result(info)
                    results.append(json)

            for f in c.functions_and_modifiers_declared:
                for node in f.nodes:
                    if node.expression and InvalidUnaryExpressionDetector(node.expression).result:
                        info = [node.function, " uses an dangerous unary operator: ", node, "\n"]
                        res = self.generate_result(info)
                        results.append(res)

        return results



from typing import Union, TYPE_CHECKING

from slither.core.solidity_types.array_type import ArrayType
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import RVALUE

if TYPE_CHECKING:
    from slither.slithir.variables.temporary import TemporaryVariable
    from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


class NewArray(Call, OperationWithLValue):
    def __init__(
        self,
        array_type: "ArrayType",
        lvalue: Union["TemporaryVariableSSA", "TemporaryVariable"],
    ) -> None:
        super().__init__()
        assert isinstance(array_type, ArrayType)
        self._array_type = array_type

        self._lvalue = lvalue

    @property
    def array_type(self) -> "ArrayType":
        return self._array_type

    @property
    def read(self) -> list[RVALUE]:
        return self._unroll(self.arguments)

    def __str__(self):
        args = [str(a) for a in self.arguments]
        lvalue = self.lvalue
        return f"{lvalue}({lvalue.type})  = new {self.array_type}({','.join(args)})"



from slither.core.expressions.expression import Expression


class TupleExpression(Expression):
    def __init__(self, expressions: list[Expression]) -> None:
        assert all(isinstance(x, Expression) for x in expressions if x)
        super().__init__()
        self._expressions = expressions

    @property
    def expressions(self) -> list[Expression]:
        return self._expressions

    def __str__(self) -> str:
        expressions_str = [str(e) for e in self.expressions]
        return "(" + ",".join(expressions_str) + ")"



import abc
from typing import Any, TYPE_CHECKING
from slither.core.context.context import Context
from slither.core.expressions.expression import Expression
from slither.core.variables.variable import Variable
from slither.utils.utils import unroll

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.core.cfg.node import Node


class AbstractOperation(abc.ABC):
    @property
    @abc.abstractmethod
    def read(self):
        """
        Return the list of variables READ
        """
        pass

    @property
    @abc.abstractmethod
    def used(self):
        """
        Return the list of variables used
        """
        pass


class Operation(Context, AbstractOperation):
    def __init__(self) -> None:
        super().__init__()
        self._node: Node | None = None
        self._expression: Expression | None = None

    def set_node(self, node: "Node") -> None:
        self._node = node

    @property
    def node(self) -> "Node":
        assert self._node
        return self._node

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self.node.compilation_unit

    @property
    def used(self) -> list[Variable]:
        """
        By default used is all the variables read
        """
        return self.read

    # if array inside the parameters
    @staticmethod
    def _unroll(l: list[Any]) -> list[Any]:
        return unroll(l)

    def set_expression(self, expression: Expression) -> None:
        self._expression = expression

    @property
    def expression(self) -> Expression | None:
        return self._expression



from typing import TYPE_CHECKING

from slither.core.source_mapping.source_mapping import SourceMapping

if TYPE_CHECKING:
    from slither.core.scope.scope import FileScope


class Pragma(SourceMapping):
    def __init__(self, directive: list[str], scope: "FileScope") -> None:
        super().__init__()
        self._directive = directive
        self.scope: FileScope = scope
        self._pattern = "pragma"

    @property
    def directive(self) -> list[str]:
        """
        list(str)
        """
        return self._directive

    @property
    def version(self) -> str:
        return "".join(self.directive[1:])

    @property
    def name(self) -> str:
        return self.version

    @property
    def is_solidity_version(self) -> bool:
        if len(self._directive) > 0:
            return self._directive[0].lower() == "solidity"
        return False

    @property
    def is_abi_encoder_v2(self) -> bool:
        if len(self._directive) == 2:
            return self._directive[0] == "experimental" and self._directive[1] == "ABIEncoderV2"
        return False

    def __str__(self) -> str:
        return "pragma " + "".join(self.directive)



from typing import Any

from slither.core.expressions import (
    AssignmentOperation,
    ConditionalExpression,
    ElementaryTypeNameExpression,
    IndexAccess,
    NewArray,
    NewContract,
    UnaryOperation,
    NewElementaryType,
)
from slither.visitors.expression.expression import ExpressionVisitor
from slither.core.expressions.call_expression import CallExpression
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.literal import Literal
from slither.core.expressions.binary_operation import BinaryOperation
from slither.core.expressions.expression import Expression
from slither.core.expressions.member_access import MemberAccess
from slither.core.expressions.tuple_expression import TupleExpression
from slither.core.expressions.type_conversion import TypeConversion


key = "ExportValues"


def get(expression: Expression) -> list[Any]:
    val = expression.context[key]
    # we delete the item to reduce memory use
    del expression.context[key]
    return val


def set_val(expression: Expression, val: list[Any]) -> None:
    expression.context[key] = val


class ExportValues(ExpressionVisitor):
    def __init__(self, expression: Expression) -> None:
        self._result: list[Expression] | None = None
        super().__init__(expression)

    def result(self) -> list[Expression]:
        if self._result is None:
            self._result = list(set(get(self.expression)))
        return self._result

    def _post_assignement_operation(self, expression: AssignmentOperation) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = left + right
        set_val(expression, val)

    def _post_binary_operation(self, expression: BinaryOperation) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = left + right
        set_val(expression, val)

    def _post_call_expression(self, expression: CallExpression) -> None:
        called = get(expression.called)
        args = [get(a) for a in expression.arguments if a]
        args = [item for sublist in args for item in sublist]
        val = called + args
        set_val(expression, val)

    def _post_conditional_expression(self, expression: ConditionalExpression) -> None:
        if_expr = get(expression.if_expression)
        else_expr = get(expression.else_expression)
        then_expr = get(expression.then_expression)
        val = if_expr + else_expr + then_expr
        set_val(expression, val)

    def _post_elementary_type_name_expression(
        self, expression: ElementaryTypeNameExpression
    ) -> None:
        set_val(expression, [])

    def _post_identifier(self, expression: Identifier) -> None:
        set_val(expression, [expression.value])

    def _post_index_access(self, expression: IndexAccess) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = left + right
        set_val(expression, val)

    def _post_literal(self, expression: Literal) -> None:
        set_val(expression, [])

    def _post_member_access(self, expression: MemberAccess) -> None:
        expr = get(expression.expression)
        val = expr
        set_val(expression, val)

    def _post_new_array(self, expression: NewArray) -> None:
        set_val(expression, [])

    def _post_new_contract(self, expression: NewContract) -> None:
        set_val(expression, [])

    def _post_new_elementary_type(self, expression: NewElementaryType) -> None:
        set_val(expression, [])

    def _post_tuple_expression(self, expression: TupleExpression) -> None:
        expressions = [get(e) for e in expression.expressions if e]
        val = [item for sublist in expressions for item in sublist]
        set_val(expression, val)

    def _post_type_conversion(self, expression: TypeConversion) -> None:
        expr = get(expression.expression)
        val = expr
        set_val(expression, val)

    def _post_unary_operation(self, expression: UnaryOperation) -> None:
        expr = get(expression.expression)
        val = expr
        set_val(expression, val)



from slither.solc_parsing.variables.variable_declaration import VariableDeclarationSolc
from slither.core.variables.event_variable import EventVariable


class EventVariableSolc(VariableDeclarationSolc):
    def __init__(self, variable: EventVariable, variable_data: dict):
        super().__init__(variable, variable_data)

    @property
    def underlying_variable(self) -> EventVariable:
        # Todo: Not sure how to overcome this with mypy
        assert isinstance(self._variable, EventVariable)
        return self._variable

    def _analyze_variable_attributes(self, attributes: dict) -> None:
        """
        Analyze event variable attributes
        :param attributes: The event variable attributes to parse.
        :return: None
        """

        # Check for the indexed attribute
        if "indexed" in attributes:
            self.underlying_variable.indexed = attributes["indexed"]

        super()._analyze_variable_attributes(attributes)



"""
Module detecting re-used base constructors in inheritance hierarchy.
"""

from typing import Any
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    ALL_SOLC_VERSIONS_04,
    DETECTOR_INFO,
)
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output


# Helper: adds explicitly called constructors with arguments to the results lookup.
def _add_constructors_with_args(
    base_constructors: list[Any | FunctionContract],
    called_by_constructor: bool,
    current_contract: Contract,
    results: dict[FunctionContract, list[tuple[Contract, bool]]],
) -> None:
    for explicit_base_constructor in base_constructors:
        if len(explicit_base_constructor.parameters) > 0:
            if explicit_base_constructor not in results:
                results[explicit_base_constructor] = []
            results[explicit_base_constructor] += [(current_contract, called_by_constructor)]


class ReusedBaseConstructor(AbstractDetector):
    """
    Re-used base constructors
    """

    ARGUMENT = "reused-constructor"
    HELP = "Reused base constructor"
    IMPACT = DetectorClassification.MEDIUM
    # The confidence is medium, because prior Solidity 0.4.22, we can't differentiate
    # contract C is A() {
    # to
    # contract C is A {
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#reused-base-constructors"

    WIKI_TITLE = "Reused base constructors"
    WIKI_DESCRIPTION = "Detects if the same base constructor is called with arguments from two different locations in the same inheritance hierarchy."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
pragma solidity ^0.4.0;

contract A{
    uint num = 5;
    constructor(uint x) public{
        num += x;
    }
}

contract B is A{
    constructor() A(2) public { /* ... */ }
}

contract C is A {
    constructor() A(3) public { /* ... */ }
}

contract D is B, C {
    constructor() public { /* ... */ }
}

contract E is B {
    constructor() A(1) public { /* ... */ }
}
```
The constructor of `A` is called multiple times in `D` and `E`:
- `D` inherits from `B` and `C`, both of which construct `A`.
- `E` only inherits from `B`, but `B` and `E` construct `A`.
."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove the duplicate constructor call."

    VULNERABLE_SOLC_VERSIONS = ALL_SOLC_VERSIONS_04

    def _detect_explicitly_called_base_constructors(
        self, contract: Contract
    ) -> dict[FunctionContract, list[tuple[Contract, bool]]]:
        """
        Detects explicitly calls to base constructors with arguments in the inheritance hierarchy.
        :param contract: The contract to detect explicit calls to a base constructor with arguments to.
        :return: Dictionary of function:list(tuple): { constructor : [(invoking_contract, called_by_constructor]}
        """
        results = {}

        # Create a set to track all completed contracts
        processed_contracts = set()
        queued_contracts = [contract] + contract.inheritance

        # Loop until there are no queued contracts left.
        while len(queued_contracts) > 0:
            # Pop a contract off the front of the queue, if it has already been processed, we stop.
            current_contract = queued_contracts.pop(0)
            if current_contract in processed_contracts:
                continue

            # Add this contract to the processed contracts
            processed_contracts.add(current_contract)

            # Prior Solidity 0.4.22, the constructor would appear two times
            # Leading to several FPs
            # As the result, we might miss some TPs if the reused is due to the constructor called
            # In the contract definition
            if self.compilation_unit.solc_version >= "0.4.22":
                # Find all base constructors explicitly called from the contract definition with arguments.
                _add_constructors_with_args(
                    current_contract.explicit_base_constructor_calls,
                    False,
                    current_contract,
                    results,
                )

            # Find all base constructors explicitly called from the constructor definition with arguments.
            if current_contract.constructors_declared:
                _add_constructors_with_args(
                    current_contract.constructors_declared.explicit_base_constructor_calls,
                    True,
                    current_contract,
                    results,
                )

        return results

    def _detect(self) -> list[Output]:
        """
        Detect reused base constructors.
        :return: Returns a list of JSON results.
        """

        results = []

        # Loop for each contract
        for contract in self.contracts:
            # Detect all locations which all underlying base constructors with arguments were called from.
            called_base_constructors = self._detect_explicitly_called_base_constructors(contract)
            for base_constructor, call_list in called_base_constructors.items():
                # Only report if there are multiple calls to the same base constructor.
                if len(call_list) <= 1:
                    continue

                # Generate data to output.
                info: DETECTOR_INFO = [
                    contract,
                    " gives base constructor ",
                    base_constructor,
                    " arguments more than once in inheritance hierarchy:\n",
                ]

                for calling_contract, called_by_constructor in call_list:
                    info += [
                        "\t- From ",
                        calling_contract,
                        f" {'constructor' if called_by_constructor else 'contract'} definition\n",
                    ]

                res = self.generate_result(info)
                results.append(res)

        return results



"""
This expression does nothing, if a contract used it, its probably a bug
"""

from slither.core.expressions.expression import Expression
from slither.core.solidity_types.type import Type
from slither.core.solidity_types.elementary_type import ElementaryType


class ElementaryTypeNameExpression(Expression):
    def __init__(self, t: ElementaryType) -> None:
        assert isinstance(t, Type)
        super().__init__()
        self._type = t

    @property
    def type(self) -> Type:
        return self._type

    @type.setter
    def type(self, new_type: Type):
        assert isinstance(new_type, Type)
        self._type = new_type

    def __str__(self) -> str:
        return str(self._type)



from .examples.backdoor import Backdoor
from .variables.uninitialized_state_variables import UninitializedStateVarsDetection
from .variables.uninitialized_storage_variables import UninitializedStorageVars
from .variables.uninitialized_local_variables import UninitializedLocalVars
from .variables.var_read_using_this import VarReadUsingThis
from .attributes.constant_pragma import ConstantPragma
from .attributes.incorrect_solc import IncorrectSolc
from .attributes.locked_ether import LockedEther
from .functions.arbitrary_send_eth import ArbitrarySendEth
from .erc.erc20.arbitrary_send_erc20_no_permit import ArbitrarySendErc20NoPermit
from .erc.erc20.arbitrary_send_erc20_permit import ArbitrarySendErc20Permit
from .functions.suicidal import Suicidal

# from .functions.complex_function import ComplexFunction
from .reentrancy.reentrancy_benign import ReentrancyBenign
from .reentrancy.reentrancy_read_before_write import ReentrancyReadBeforeWritten
from .reentrancy.reentrancy_eth import ReentrancyEth
from .reentrancy.reentrancy_no_gas import ReentrancyNoGas
from .reentrancy.reentrancy_events import ReentrancyEvent
from .reentrancy.reentrancy_balance import ReentrancyBalance
from .variables.unused_state_variables import UnusedStateVars
from .variables.could_be_constant import CouldBeConstant
from .variables.could_be_immutable import CouldBeImmutable
from .statements.tx_origin import TxOrigin
from .statements.assembly import Assembly
from .operations.low_level_calls import LowLevelCalls
from .operations.unused_return_values import UnusedReturnValues
from .operations.unchecked_transfer import UncheckedTransfer
from .naming_convention.naming_convention import NamingConvention
from .functions.external_function import ExternalFunction
from .statements.controlled_delegatecall import ControlledDelegateCall
from .attributes.const_functions_asm import ConstantFunctionsAsm
from .attributes.const_functions_state import ConstantFunctionsState
from .shadowing.abstract import ShadowingAbstractDetection
from .shadowing.state import StateShadowing
from .shadowing.local import LocalShadowing
from .shadowing.builtin_symbols import BuiltinSymbolShadowing
from .operations.block_timestamp import Timestamp
from .statements.calls_in_loop import MultipleCallsInLoop
from .statements.incorrect_strict_equality import IncorrectStrictEquality
from .erc.erc20.incorrect_erc20_interface import IncorrectERC20InterfaceDetection
from .erc.incorrect_erc721_interface import IncorrectERC721InterfaceDetection
from .erc.unindexed_event_parameters import UnindexedERC20EventParameters
from .statements.deprecated_calls import DeprecatedStandards
from .source.rtlo import RightToLeftOverride
from .statements.too_many_digits import TooManyDigits
from .operations.unchecked_low_level_return_values import UncheckedLowLevel
from .operations.unchecked_send_return_value import UncheckedSend
from .operations.void_constructor import VoidConstructor
from .statements.type_based_tautology import TypeBasedTautology
from .statements.boolean_constant_equality import BooleanEquality
from .statements.boolean_constant_misuse import BooleanConstantMisuse
from .statements.divide_before_multiply import DivideBeforeMultiply
from .statements.unprotected_upgradeable import UnprotectedUpgradeable
from .slither.name_reused import NameReused

from .functions.unimplemented import UnimplementedFunctionDetection
from .statements.mapping_deletion import MappingDeletionDetection
from .statements.array_length_assignment import ArrayLengthAssignment
from .variables.function_init_state_variables import FunctionInitializedState
from .statements.redundant_statements import RedundantStatements
from .operations.bad_prng import BadPRNG
from .statements.costly_operations_in_loop import CostlyOperationsInLoop
from .statements.assert_state_change import AssertStateChange
from .attributes.unimplemented_interface import MissingInheritance
from .assembly.shift_parameter_mixup import ShiftParameterMixup
from .compiler_bugs.storage_signed_integer_array import StorageSignedIntegerArray
from .compiler_bugs.uninitialized_function_ptr_in_constructor import (
    UninitializedFunctionPtrsConstructor,
)
from .compiler_bugs.storage_ABIEncoderV2_array import ABIEncoderV2Array
from .compiler_bugs.array_by_reference import ArrayByReference
from .compiler_bugs.enum_conversion import EnumConversion
from .compiler_bugs.multiple_constructor_schemes import MultipleConstructorSchemes
from .compiler_bugs.public_mapping_nested import PublicMappingNested
from .compiler_bugs.reused_base_constructor import ReusedBaseConstructor
from .operations.missing_events_access_control import MissingEventsAccessControl
from .operations.missing_events_arithmetic import MissingEventsArithmetic
from .functions.modifier import ModifierDefaultDetection
from .variables.predeclaration_usage_local import PredeclarationUsageLocal
from .statements.unary import IncorrectUnaryExpressionDetection
from .operations.missing_zero_address_validation import MissingZeroAddressValidation
from .functions.dead_code import DeadCode
from .statements.write_after_write import WriteAfterWrite
from .statements.msg_value_in_loop import MsgValueInLoop
from .statements.delegatecall_in_loop import DelegatecallInLoop
from .functions.protected_variable import ProtectedVariables
from .functions.permit_domain_signature_collision import DomainSeparatorCollision
from .functions.cyclomatic_complexity import CyclomaticComplexity
from .operations.cache_array_length import CacheArrayLength
from .statements.incorrect_using_for import IncorrectUsingFor
from .operations.encode_packed import EncodePackedCollision
from .assembly.incorrect_return import IncorrectReturn
from .assembly.return_instead_of_leave import ReturnInsteadOfLeave
from .operations.incorrect_exp import IncorrectOperatorExponentiation
from .statements.tautological_compare import TautologicalCompare
from .statements.return_bomb import ReturnBomb
from .functions.out_of_order_retryable import OutOfOrderRetryable
from .functions.gelato_unprotected_randomness import GelatoUnprotectedRandomness
from .statements.chronicle_unchecked_price import ChronicleUncheckedPrice
from .statements.pyth_unchecked_confidence import PythUncheckedConfidence
from .statements.pyth_unchecked_publishtime import PythUncheckedPublishTime
from .functions.chainlink_feed_registry import ChainlinkFeedRegistry
from .functions.pyth_deprecated_functions import PythDeprecatedFunctions
from .functions.optimism_deprecation import OptimismDeprecation
from .attributes.unindexed_event_address import UnindexedEventAddress

# from .statements.unused_import import UnusedImport
